from typing import Callable, Dict, List, Optional, Union

from fastapi import FastAPI
from fastapi.routing import Mount
import numpy as np

from furiosa.runtime.tensor import TensorDesc
from furiosa.server import ModelConfig, NuxModel


class ServeModel:
    def __init__(
        self,
        app: FastAPI,
        name: str,
        *,
        model: Union[str, bytes],
        version: Optional[str] = None,
        description: Optional[str] = None,
        npu_device: Optional[str] = None,
        compiler_config: Optional[Dict] = None,
    ):
        self._app = app
        self._config = ModelConfig(
            name=name,
            model=model,
            version=version,
            description=description,
            npu_device=npu_device,
            compiler_config=compiler_config,
        )

        self._model = NuxModel(self._config)
        self._routes: Dict[Callable, Callable] = {}

    def expose(self):
        """
        Expose FastAPI route API endpoint
        """
        for func, decorator in self._routes.items():
            # Decorate the path operation function to expose endpoint
            decorator(func)

    def hide(self):
        """
        Hide FastAPI route API endpoint
        """
        # Gather routes not in sub applications
        routes = [route for route in self._app.routes if not isinstance(route, Mount)]

        # Target routes to be removed
        targets = [route for route in routes if route.endpoint in self._routes]  # type: ignore

        # Unregister path operation functions to hide endpoint
        for route in targets:
            self._app.routes.remove(route)

    async def predict(self, payload: List[np.ndarray]) -> List[np.ndarray]:
        return await self._model.predict(payload)

    @property
    def inner(self) -> NuxModel:
        return self._model

    @property
    def config(self) -> ModelConfig:
        return self._config

    @property
    def inputs(self) -> List[TensorDesc]:
        return self._model.session.inputs()

    @property
    def outputs(self) -> List[TensorDesc]:
        return self._model.session.outputs()

    def _method(self, kind: str, *args, **kwargs) -> Callable:
        def decorator(func):
            """
            Register FastAPI path operation function to be used later.

            The function will be registerd into FastAPI app when model is loaded.
            """
            self._routes[func] = getattr(self._app, kind)(*args, **kwargs)
            return func

        return decorator

    def get(self, *args, **kwargs) -> Callable:
        return self._method("get", *args, **kwargs)

    def put(self, *args, **kwargs) -> Callable:
        return self._method("put", *args, **kwargs)

    def post(self, *args, **kwargs) -> Callable:
        return self._method("post", *args, **kwargs)

    def delete(self, *args, **kwargs) -> Callable:
        return self._method("delete", *args, **kwargs)

    def head(self, *args, **kwargs) -> Callable:
        return self._method("head", *args, **kwargs)

    def patch(self, *args, **kwargs) -> Callable:
        return self._method("patch", *args, **kwargs)

    def trace(self, *args, **kwargs) -> Callable:
        return self._method("trace", *args, **kwargs)



from abc import ABC, abstractmethod
from functools import wraps
from typing import Any, Callable


class Processor(ABC):
    @abstractmethod
    async def preprocess(self, *args: Any, **kwargs: Any) -> Any:
        ...

    @abstractmethod
    async def postprocess(self, *args: Any, **kwargs: Any) -> Any:
        ...

    def __call__(self, func: Callable):
        """
        Rerturn decorator which will call preprocess and postprocess.

        Note that the function signatures (preprocess, infer, postprocess) must be
        compatible to make the pipelines correctly.
        """

        @wraps(
            self.preprocess,
            # Note that we don't have any assigned value. See below
            assigned=(),
        )
        async def decorator(*args: Any, **kwargs: Any) -> Any:
            # Preprocess
            output = await self.preprocess(*args, **kwargs)
            # Infer
            response = await func(output)
            # Postprocess
            return await self.postprocess(response)

        # # Extract parameter annotation from preprocess
        params = {
            key: value for key, value in self.preprocess.__annotations__.items() if key != "return"
        }
        # Extract return annotation from postprocess
        returns = {"return": self.postprocess.__annotations__["return"]}

        # Replace original function's annotation with new signature
        decorator.__annotations__ = dict(params, **returns)

        # Assign original attribute to decorator. Do not use wrap() as FastAPI will unwrap it
        for attr in ("__module__", "__name__", "__qualname__"):
            setattr(decorator, attr, getattr(func, attr))

        # Integrate docstring
        decorator.__doc__ = "".join(
            (
                doc
                for doc in (
                    self.preprocess.__doc__,
                    func.__doc__,
                    self.postprocess.__doc__,
                )
                if doc
            )
        )

        return decorator



from fastapi import FastAPI
from fastapi.routing import APIRoute

from furiosa.server import ModelServerError
from furiosa.server.api.rest.app import exception_handler
from furiosa.server.api.rest.endpoints import ModelEndpoints
from furiosa.server.handlers import PredictHandler
from furiosa.server.settings import ServerConfig

from .repository import repository

endpoints = ModelEndpoints(
    PredictHandler(
        # TODO(yan): Accept user configs
        ServerConfig(),
        repository,
    )
)

app = FastAPI(
    routes=[
        APIRoute(
            "/{model_name}/ready",
            endpoints.model_ready,
        ),
        APIRoute(
            "/{model_name}/versions/{model_version}/ready",
            endpoints.model_ready,
        ),
        # Model metadata
        APIRoute(
            "/{model_name}",
            endpoints.model_metadata,
        ),
        APIRoute(
            "/{model_name}/versions/{model_version}",
            endpoints.model_metadata,
        ),
    ],
    exception_handlers={ModelServerError: exception_handler},  # type: ignore
)



import os
from typing import Any, Dict, Optional

from fastapi import FastAPI

from furiosa.registry.client import transport
from furiosa.server.model import Model
from furiosa.server.registry import InMemoryRegistry
from furiosa.server.repository import Repository

from .apps.repository import repository
from .model import ServeModel


class ServeAPI:
    def __init__(
        self,
        repository: Repository = repository,
        **kwargs: Any,
    ):
        self._app = FastAPI(**kwargs, on_startup=[self.load])
        self._models: Dict[Model, ServeModel] = {}
        self._repository = repository
        self._registry = next(
            (
                registry
                for registry in repository.registries
                if isinstance(registry, InMemoryRegistry)
            ),
            None,
        )

        assert (
            self._registry is not None
        ), "InMemoryRegistry is required to use Repository in ServeAPI"

        self._repository.on_load = self._on_load
        self._repository.on_unload = self._on_unload

    @property
    def app(self) -> FastAPI:
        return self._app

    async def model(
        self,
        name: str,
        *,
        location: str,
        version: Optional[str] = None,
        description: Optional[str] = None,
        npu_device: Optional[str] = None,
        compiler_config: Optional[Dict] = None,
    ):
        # Assume that relative path is only for file
        if transport.is_relative(location):
            location = os.path.join("file://", location)

        model = ServeModel(
            app=self._app,
            name=name,
            model=await transport.read(location),
            version=version,
            description=description,
            npu_device=npu_device,
            compiler_config=compiler_config,
        )

        self._models[model.inner] = model

        # Register model config for model discovery
        assert self._registry is not None
        self._registry.register(model.config)

        return model

    async def load(self):
        print("application load")
        for inner, serve_model in self._models.items():
            # Load models via repository
            await self._repository.load(inner)

    def _on_load(self, model: Model):
        """
        Callback function which expose API endpoints when model is loaded
        """
        for inner, serve_model in self._models.items():
            # (name, version) pair gurantees Model identity
            # TODO(yan): Refactor repository API to load ServeModel directly
            if inner.name == model.name and inner.version == model.version:
                serve_model.expose()

    def _on_unload(self, inner: Model):
        """
        Callback function which hide API endpoints when model is unloaded
        """
        serve_model = self._models[inner]
        serve_model.hide()



from fastapi import FastAPI
from fastapi.routing import APIRoute

from furiosa.server import ModelServerError
from furiosa.server.api.rest.app import exception_handler
from furiosa.server.api.rest.endpoints import ModelRepositoryEndpoints
from furiosa.server.handlers import RepositoryHandler
from furiosa.server.registry import InMemoryRegistry
from furiosa.server.repository import Repository

registry = InMemoryRegistry()

repository = Repository([registry])

endpoints = ModelRepositoryEndpoints(RepositoryHandler(repository))

app = FastAPI(
    routes=[
        APIRoute(
            "/index",
            endpoints.index,
            methods=["POST"],
        ),
        APIRoute(
            "/models/{model_name}/load",
            endpoints.load,
            methods=["POST"],
        ),
        APIRoute(
            "/models/{model_name}/unload",
            endpoints.unload,
            methods=["POST"],
        ),
    ],
    exception_handlers={ModelServerError: exception_handler},  # type: ignore
)



from typing import Dict

# Not yet typed. See https://github.com/python-pillow/Pillow/issues/2625
from PIL import Image  # type: ignore
from fastapi import File, UploadFile
import numpy as np

from .. import ServeModel
from .base import Processor


class ImageNet(Processor):
    def __init__(self, model: ServeModel, label: str):
        self.model = model
        self.label = label

    async def preprocess(self, image: UploadFile = File(...)) -> np.ndarray:  # type: ignore
        """
        Preprocess to convert a image (Python file-like object) to Numpy array
        """

        # Get model input tensor to find out tesnor shape
        _, height, width, channel = self.model.inputs[0].shape

        # Convert PIL image to Numpy array
        data = np.zeros((width, height, channel), np.uint8)
        data[:width, :height, :channel] = (
            Image.open(image.file).convert("RGB").resize((width, height))
        )
        return np.reshape(data, (1, width, height, channel))

    async def postprocess(self, output: np.ndarray) -> Dict:  # type: ignore
        """
        Postprocess to classify image from compiled model with labels
        """

        classified = np.squeeze(output)

        # Load pre-defined labels
        labels = {index: line.strip() for index, line in enumerate(open(self.label).readlines())}

        # Find objects with index which mostly fits
        objects = sorted((score, index[0]) for index, score in np.ndenumerate(classified))[::-1]

        # Return fifth best fit image labels with scores. Note that we are casting int here to
        # convert numpy uin8 type into Python native int type to allow FastAPI serialize result JSON
        return {labels[index]: int(score) for score, index in objects[:5]}



from fastapi import FastAPI
from fastapi.routing import APIRoute

from furiosa.server import ModelServerError
from furiosa.server.api.rest.app import exception_handler

from .model import endpoints

app = FastAPI(
    routes=[
        # Liveness and readiness
        APIRoute("/live", endpoints.live),
        APIRoute("/ready", endpoints.ready),
        # Server metadata
        APIRoute(
            "/",
            endpoints.metadata,
        ),
    ],
    exception_handlers={ModelServerError: exception_handler},  # type: ignore
)
